; 

    .486                       ; create 32 bit code
    .model flat, stdcall       ; 32 bit memory model
    option casemap :none       ; case sensitive



    include     \masm32\include\windows.inc
    include     \masm32\include\kernel32.inc
    include     \masm32\macros\macros.asm
    include     \masm32\include\masm32.inc
    include     \masm32\include\debug.inc
    
    includelib  \masm32\lib\kernel32.lib
    includelib  \masm32\lib\debug.lib

; 
; LZW_init      PROTO :DWORD
; LZW_encode    PROTO :DWORD,:DWORD,:DWORD
; LZW_decode    PROTO :DWORD,:DWORD,:DWORD

PUBLIC  LZW_dictionary,LZW_entry_count

; 

.data?
    align 16

    LZW_entry_count dd ?                            ; number of entries in dictionary
    LZW_max_entry   dd ?
    LZW_dictionary  dd ?,?                          ; pointer to dictionary buffer start,end
    LZW_code_mask   dd ?                            ; bitmask for codewords
    LZW_buffer_end  dd ?                            ; auxillary variable
    LZW_s           dd ?
    LZW_code_size   dd ?                            ; code size in bits (12-15)

    LZW_stream_offs dd ?

; 

.code

align 4

; 
;
; Initializes variables and allocates memory for the dictionary
;
; -------------------------------------------------------------------------
;  IN:  cwSize - code word size in bits (12-16 bits)
; -------------------------------------------------------------------------

LZW_init        PROC USES eax ecx,cwSize:DWORD

                mov     ecx,cwSize
                mov     eax,1
                mov     [LZW_code_size],ecx             ; code size
                shl     eax,cl
                mov     [LZW_max_entry],eax
                shl     eax,2
                push    eax
                sub     eax,1
                mov     [LZW_code_mask],eax
                pop     eax
                invoke HeapAlloc,FUNC(GetProcessHeap),HEAP_ZERO_MEMORY,eax
                mov     [LZW_dictionary],eax
                xor     eax,eax
                mov     [LZW_stream_offs],eax           ; bit stream offset
                mov     [LZW_entry_count],eax
                ret

LZW_init        ENDP
; 
;
; Frees allocated memory and clears variables
;
; -------------------------------------------------------------------------

LZW_destroy     PROC USES eax

                invoke HeapFree,FUNC(GetProcessHeap),0,[LZW_dictionary]
                xor     eax,eax
                mov     [LZW_entry_count],eax
                mov     [LZW_dictionary],eax
                ret

LZW_destroy     ENDP
; 
;
; Encodes compressed data according to the LZW method
;
; -------------------------------------------------------------------------
;  IN:    src - pointer to input data
;         dst - pointer to output data
;       bytes - count of input bytes
; -------------------------------------------------------------------------
LZW_encode      PROC USES ebx ecx edx esi edi,src:DWORD,dst:DWORD,bytes:DWORD

;PrintText '==ENCODE=============='
                mov     esi,src
                mov     edi,dst

                add     bytes,esi

                xor     eax,eax
                mov     al,[esi]                ; Read first char
; -------------------------------------------------------------------------
; Main loop
; ---------------
@_1:
                mov     ecx,[LZW_code_size]
                add     esi,1
               
                shl     eax,cl

                mov     ecx,[LZW_entry_count]
                mov     al,[esi]
;PrintHex eax,'read'
; -------------------------------------------------------------------------
; LZW_lookup
; ---------------
;PrintHex eax,'lookup'
                push    esi
                mov     esi,[LZW_dictionary]

                sub     ecx,1
                jc  @_10
@_LU1:
                cmp     eax,[esi+ecx*4]
                jz  @_2

                sub     ecx,1
                jnc @_LU1
; -------------------------------------------------------------------------
; Not found
; ---------------
@_10:
;PrintHex eax,'not found'
; -------------------------------------------------------------------------
; LZW_add_entry
; ---------------
                mov     ebx,[LZW_entry_count]
                mov     esi,[LZW_dictionary]

                cmp     ebx,[LZW_max_entry]
                jnc  @_11

;PrintHex eax,'new entry'
                mov     [esi+ebx*4],eax
                add     [LZW_entry_count],1
; ---------------
@_11:
; -------------------------------------------------------------------------
; LZW_write_code
; ---------------
                pop     esi
                mov     ecx,[LZW_code_size]

                shr     eax,cl

                mov     ebx,[edi]
                mov     ecx,[LZW_stream_offs]

                shl     eax,cl                       ; shift the code

                add     ecx,[LZW_code_size]          ; offset increment in bits
                or      eax,ebx

                mov     bl,cl
                mov     [edi],eax                   ; write the code, buffer must originally be filled with zero

                shr     cl,3                        ; /8
                and     ebx,07h                     ; new offset in bits

                add     edi,ecx                     ; increment in bytes
                mov     [LZW_stream_offs],ebx
; -------------------------------------------------------------------------
                xor     eax,eax
                mov     al,[esi]

;PrintHex eax,'code'
                jmp @_3
; -------------------------------------------------------------------------
; Found:
; ---------------
@_2:
;PrintHex eax,'found'
                add     ecx,0100h
                pop     esi

                mov     eax,ecx
; ---------------
@_3:
;PrintText '--------'
                cmp     esi,bytes
                jc     @_1
; -------------------------------------------------------------------------
                or      bl,bl
                jz  @_31

                add     edi,1
@_31:
                mov     eax,edi
                sub     eax,dst

                ret
                
LZW_encode      ENDP
; 
end